from distutils.dir_util import copy_tree
import shutil
import glob
import os

def init_input_folder(self) -> bool:
    """
    Creating the tree structure for temporary files storage.
    By default it will be located in /tmp/frelatage
    """
    start_input_directory = self.input_directory
    tmp_input_directory = self.config.FRELATAGE_INPUT_FILE_TMP_DIR

    # Create /tmp/frelatage (default value) directory if not exists
    if not os.path.exists(tmp_input_directory):
        os.makedirs(tmp_input_directory)

    # Empty the directory
    else:
        list_dir = os.listdir(tmp_input_directory)
        for filename in list_dir:
            file_path = os.path.join(tmp_input_directory, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)

    # Create a subdirectory /tmp/frelatage/<subfolder> for each thread
    for thread in range(self.threads_count):
        thread_directory = os.path.join(tmp_input_directory, str(thread))
        os.makedirs(thread_directory)

        # Files located in the input folder that are used in this stage
        filenames = [file_argument.value for file_argument in filter(lambda a: a.file, self.queue.current_arguments())]
        # Copy the useful files into the subdirectory
        for i in range(len(filenames)):
            # /tmp/frelatage/<thread>/<argument position>
            argument_directory = os.path.join(thread_directory, str(i))
            os.makedirs(argument_directory)
            # ./in/<filename>
            input_file = os.path.join(start_input_directory, filenames[i])
            # /tmp/frelatage/<thread>/<argument position>/<filename>
            output_file = os.path.join(argument_directory, filenames[i])
            shutil.copyfile(input_file, output_file)
    return True

def init_file_input_arguments(self) -> bool:
    """
    Initialization of the file arguments values.

    Input {
        type: File
        value: <input_file_tmp_dir>/<thread 0>/<argument position>/<filename>
    }
    """
    file_input_arguments_count = 0
    # We initialize the value of each argument of type "file".
    for i in range(len(self.arguments)):
        if self.arguments[i].file:
            # <input_file_tmp_dir>/<thread 0>/<argument position>/<filename>
            self.arguments[i].value = "{base_directory}/{thread}/{position}/{filename}".format(
                base_directory=self.config.FRELATAGE_INPUT_FILE_TMP_DIR,
                thread="0",
                position=str(i), 
                filename=self.arguments[i].value
            )
            file_input_arguments_count+= 1
    return True

def init_file_inputs(self) -> bool:
    """
    Set up the tree structure to fuzz a function with "file" type arguments
    """
    if any([argument.file for argument in self.arguments]):
        self.init_input_folder()
        self.init_file_input_arguments()


class Input:
    """
    Input can be either a file or any other type
    """
    def __init__(self, value: str = "", file: bool = False) -> None:
        self.value = value
        self.file = file
    
    def __iter__(self) -> None:
        yield 'value', self.value
        yield 'file', self.file


from frelatage.corpus.corpus import load_corpus
from frelatage.config.config import Config
from frelatage.tracer.tracer import Tracer
from frelatage.mutator.mutator import *
from frelatage.input.input import Input 
from frelatage.queue.queue import Queue
from typing import Type, Iterable, Callable, List
from importlib.metadata import version
from datetime import datetime
import time
import sys
import os

# Automatically retrieves the current version number
__version__ = version('frelatage')

class Fuzzer(object):
    """
    Copyright (c) 2022 Rog3rSm1th

    Frelatage is a coverage-based Python fuzzing library which can be used to fuzz python code. 
    The development of Frelatage was inspired by various other fuzzers, including AFL/AFL++, 
    Atheris and PyFuzzer.The main purpose of the project is to take advantage of the best features 
    of these fuzzers and gather them together into a new tool in order to efficiently fuzz python applications.
    """
    from ._mutation import valid_mutators, get_mutation, generate_cycle_mutations
    from ._interface import init_interface, refresh_interface, start_interface, exit_message
    from ._evaluate import evaluate_mutations
    from ._cycle import run_function, run_cycle  
    from ._fuzz import fuzz
    from ._report import get_report_name, save_report
    from ._input import init_input_folder, init_file_input_arguments, init_file_inputs

    def __init__(self,
                 method: Callable,
                 corpus: list[object],
                 threads_count: int = 8,
                 exceptions_whitelist: list = (),
                 exceptions_blacklist: list = (),
                 output_directory: str = "./out",
                 silent: bool = False
        ) -> None:
        """
        Initialize the fuzzer
        """        
        # Frelatage configuration
        self.config = Config

        # Global set of reached instructions
        self.reached_instructions= set([])
        # Global set of instructions pairs
        self.instructions_pairs = set([])
        # Global set of instruction pairs executed during a crash
        self.favored_pairs = set([])
        # Global Set of positions where a crash occurred
        self.error_positions = set([])
        # Fuzzed method
        self.method = method

        # Remove duplicates in coprus
        corpus = [list(set(argument)) for argument in corpus]
        # Frelatage corpus
        self.corpus = corpus
        # List of all avalaibles mutators
        self.mutators = mutators
        # Number of concurrently launched threads
        self.threads_count = max(min(threads_count, Config.FRELATAGE_MAX_THREADS), 8)
        # List of cycle mutations
        self.cycle = []

        # Exceptions that will be taken into account or not when fuzzing
        self.exceptions_whitelist = exceptions_whitelist
        self.exceptions_blacklist = exceptions_blacklist

        # Frelatage tracer
        self.tracer = Tracer(exceptions_whitelist=self.exceptions_whitelist, exceptions_blacklist=self.exceptions_blacklist)

        # Input and output directories
        # The working directory is the same as the fuzz file
        self.input_directory = os.path.join(
            os.path.dirname(os.path.realpath(sys.argv[0])),
            Config.FRELATAGE_INPUT_DIR
        )
        self.output_directory = os.path.join(
            os.path.dirname(os.path.realpath(sys.argv[0])),
            output_directory
        )
        
        # Silent output
        self.silent = silent

        # Fuzzer statistics
        self.cycles_count = 0
        self.inputs_count = 0
        self.stage_inputs_count = 0
        self.unique_crashes = 0
        self.total_crashes = 0
        self.unique_timeout = 0
        self.total_timeouts = 0
        
        # Time statistics
        self.fuzz_start_time = datetime.now()
        self.last_new_path_time = None
        self.last_unique_crash_time = None
        self.last_unique_timeout_time = None

        # Genetic algorithm parameters
        self.survival_probability = 0.5
        self.mutation_probability = 0.3

         # Number of Frelatage cycles without finding new paths
        self.cycles_without_new_path = 0
        # Corpus entries that are still in the queue
        self.queue= Queue(self.corpus)
        # Current arguments
        self.arguments = self.queue.current_arguments()

        # Initialize file input folders in /tmp/frelatage (default value)
        # Can be modified using the FRELATAGE_INPUT_FILE_TMP_DIR env variable
        self.init_file_inputs()


from datetime import datetime

def evaluate_mutations(self, reports: list) -> list:
    """
    Evaluates input reports and determines the number of unique bugs as well
    as the number of unique instruction bugs.
    Return the n mutations with the best score.
    """
    start_reached_instructions_count = len(self.reached_instructions)

    # Identify uniques crashes
    for report in reports:
        mutation_reached_instructions = set(report.trace.reached_instructions)
        mutation_instructions_pairs = set(report.trace.instructions_pairs)

        if report.error:
            self.total_crashes += 1
        if report.timeout:
            self.total_timeouts += 1

        # A crash  is unique if :
        # -> An error/timeout is triggered AND a new pair of instruction leading to a crash is found
        # OR
        # -> An error/timeout is triggered at a new position
        if report.error or report.timeout:

            new_error = False
            # We check if there are new instructions pairs
            if (report.new_error_instruction is not None) and (report.trace.error_position not in self.error_positions):
                new_error = True
                # We update the list of instructions that led to a crash
                self.error_positions = set.union(self.error_positions, [report.trace.error_position]) 

            is_unique = (len(set.union(mutation_instructions_pairs, self.favored_pairs)) != len(self.favored_pairs)) or new_error
            # Unique bugs
            if is_unique:
                self.favored_pairs = set.union(self.favored_pairs, mutation_instructions_pairs)  
                self.last_unique_crash_time = datetime.now()
                self.save_report(report)
                self.unique_crashes += 1
                # Unique timeouts
                if report.timeout:
                    self.unique_timeout += 1
                    self.last_unique_timeout = datetime.now()

        # Update the list of global reached instructions and global instruction pairs
        self.reached_instructions = set.union(self.reached_instructions, mutation_reached_instructions)
        self.instructions_pairs = set.union(self.instructions_pairs, mutation_instructions_pairs)   

    # If we found a new path
    if len(self.reached_instructions) > start_reached_instructions_count:
        self.last_new_path_time = datetime.now()
        self.cycles_without_new_path = 0
    else:
        self.cycles_without_new_path += 1

    # Rank reports
    ranked_reports = sorted(reports, reverse=True)

    # Select the best mutation
    best_mutations_count = int(self.threads_count * self.survival_probability)
    best_mutations = [mutation.input for mutation in ranked_reports[:best_mutations_count]]
    return best_mutations


from frelatage.config.config import Config
from threading import Thread
import os

def fuzz(self) -> None:
    """
    Run the fuzzer
    """
    try:
        # Interface
        if not self.silent:
            p = Thread(target=self.start_interface)
            p.daemon = True
            p.start()

        # Fuzzing
        parents = [self.arguments]
        while True: 
            self.generate_cycle_mutations(parents)
            reports = self.run_cycle()
            # If no new paths have been found for a while, we go to the next stage
            if self.cycles_without_new_path >= Config.FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS:
                # Next stage
                self.queue.position += 1
                if not self.queue.end:
                    # Initialize the new stage
                    self.arguments = self.queue.current_arguments()
                    parents = [self.arguments]
                    self.cycles_without_new_path = 0
                    self.stage_inputs_count = 0
                    self.init_file_inputs()
                # End of the fuzzing process
                # Exit the program
                else:
                    self.exit_message(normal_ending=True)
                    exit(1)
            else:
                parents = self.evaluate_mutations(reports)
    # Exit the program
    # Keyboard interrupt
    except KeyboardInterrupt:
        self.exit_message(aborted_by_user=True)
        exit(0)
    # Error in Frelatage
    except Exception:
        self.exit_message(aborted_by_user=False)
        exit(1)


from frelatage.colors import Colors
from frelatage import __version__
from datetime import datetime
from string import Formatter
from curses import wrapper
import curses
import time
import os

# Refresh the interface 10 times/second
REFRESH_INTERVAL = 0.1

def format_delta(time_delta:datetime, format: str) -> str:
    """
    Format a time delta.
    """
    formatter = Formatter()
    digits = {}
    constants = {'D': 86400, 'H': 3600, 'M': 60, 'S': 1}
    k = map(lambda x: x[1], list(formatter.parse(format)))
    remaining = int(time_delta.total_seconds())

    for i in ('D', 'H', 'M', 'S'):
        if i in k and i in constants.keys():
            digits[i], remaining = divmod(remaining, constants[i])

    return formatter.format(format, **digits)

def format_time_elapsed(datetime: datetime) -> str:
    """
    Format the time elapsed since a datetime in the format : 
    {D} days, {H} hrs, {M} min, {S} sec
    """
    if datetime is not None:
        format_time_delta = datetime.now() - datetime
        format_time = format_delta(format_time_delta, "{D} days, {H} hrs, {M} min, {S} sec")
    # If the event has not yet occurred
    else:
        format_time = "none seen yet" 
    return format_time

def init_interface(self, stdscr) -> bool:
    """
    Initialize the Frelatage CLI
    """
    # Hide cursor
    curses.curs_set(0)
    # Colors
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
    stdscr.bkgd(' ', curses.color_pair(1) | curses.A_BOLD)

    self.screen = stdscr
    self.screen.clear()
    self.screen.refresh()
    return True

def refresh_interface(self):
    """
    Refresh the Frelatage CLI
    """
    # Title
    title = "Frelatage {version} ({function_name})".format(
        version=__version__,
        function_name=self.method.__name__
    ).center(105)

    # Process timing
    run_time = format_time_elapsed(self.fuzz_start_time).ljust(32)
    last_new_path_time = format_time_elapsed(self.last_new_path_time).ljust(32)
    last_unique_crash_time = format_time_elapsed(self.last_unique_crash_time).ljust(32)
    last_unique_timeout_time = format_time_elapsed(self.last_unique_timeout_time).ljust(32)
    
    # Overall results
    uniques_crashes_count = str(self.unique_crashes).ljust(9)
    uniques_timeouts_count = str(self.unique_timeout).ljust(9)

    # Finding in depth 
    total_paths_count = str(len(self.reached_instructions)).ljust(22)
    favored_paths_count = len(self.favored_pairs)
    favored_paths_rate = round(int(favored_paths_count)/int(total_paths_count) * 100, 2) if int(total_paths_count) else 0.00
    favored_paths = "{favored_paths} ({rate}%)".format(favored_paths=favored_paths_count, rate=favored_paths_rate).ljust(22)
    
    # Crashes
    total_crashes = "{crashes} ({uniques} uniques)".format(crashes=str(self.total_crashes), uniques=str(self.unique_crashes)).ljust(22)
    total_timeouts = "{total_timeouts}".format(total_timeouts=self.total_timeouts).ljust(22)

    # Progress
    cycles_count = str(self.cycles_count).ljust(12)
    total_executions = str(self.inputs_count).ljust(12)

    # Stage progress
    current_argument = self.queue.position + 1
    total_arguments_count = len(self.queue.arguments)
    current_stage = "{current_argument}/{total_arguments_count}".format(current_argument=current_argument, total_arguments_count=total_arguments_count).ljust(16)
    stage_executions = str(self.stage_inputs_count).ljust(16)

    # Interface
    self.screen.addstr(0, 0, """
    {title}
                                        
    +---- Process timing ------------------------------------+--- Finding in depth -------------------------+
    | Run time            :: {run_time}| Favored paths       :: {favored_paths}|
    | Last new path       :: {last_new_path_time}| Total paths         :: {total_paths_count}|
    | Last unique crash   :: {last_unique_crash_time}| Total timeouts      :: {total_timeouts}|
    | Last unique timeout :: {last_unique_timeout_time}| Total crashes       :: {total_crashes}|
    +---- Overall result -------------+---- Global progress -+-------------+---- Stage progress-------------+
    | Uniques crashes     :: {uniques_crashes_count}| Cycles done         :: {cycles_count}| Stage       :: {current_stage}|
    | Unique timeouts     :: {uniques_timeouts_count}| Total executions    :: {total_executions}| Stage execs :: {stage_executions}|
    +---------------------------------+------------------------------------+--------------------------------+
    """.format(
            title=title,
            run_time=run_time,
            last_new_path_time=last_new_path_time,
            last_unique_crash_time=last_unique_crash_time,
            last_unique_timeout_time=last_unique_timeout_time,
            uniques_crashes_count=uniques_crashes_count,
            uniques_timeouts_count=uniques_timeouts_count,
            favored_paths=favored_paths,
            total_paths_count=total_paths_count,
            total_timeouts=total_timeouts,
            total_crashes=total_crashes,
            cycles_count=cycles_count,
            total_executions=total_executions,
            current_stage=current_stage,
            stage_executions=stage_executions
        )
    )
    self.screen.refresh()

def exit_message(self, normal_ending: bool = False, aborted_by_user: bool = False) -> bool:
    """
    Message displayed when exiting the program
    """
    run_time = format_time_elapsed(self.fuzz_start_time)
    uniques_crashes_count = str(self.unique_crashes)
    uniques_timeouts_count = str(self.unique_timeout)
    total_crashes = "{crashes} ({uniques} uniques)".format(crashes=str(self.total_crashes), uniques=str(self.unique_crashes))
    total_timeouts = "{total_timeouts}".format(total_timeouts=self.total_timeouts)
    total_paths_count = str(len(self.reached_instructions))
    cycles_count = str(self.cycles_count)
    total_executions = str(self.inputs_count)

    # End the curse window 
    if not self.silent:
        curses.endwin()

    # Keyboard interrupt
    if aborted_by_user:
        print(Colors.FAIL + "+++ Fuzzing aborted by user +++" + "\r\n")
    # Normal ending
    elif normal_ending:
        print(Colors.OKGREEN + "+++ Fuzzing completed +++" + "\r\n")
    # Error in the program
    else:
        print(Colors.FAIL + "+++ Fuzzing was interrupted by an error in Frelatage +++" + "\r\n")
    
    # Message displayed at the end of the program 
    print(Colors.OKGREEN + "[+] " + Colors.ENDC + "Error reports are located in: " + Colors.BOLD + self.output_directory + Colors.ENDC + "\r\n"
          + "\r\n"
          + Colors.BOLD + "Fuzzing statistics: " + Colors.ENDC + "\r\n"
          + "\r\n"
          + "Total run time: " + run_time + "\r\n"
          + "Uniques crashes: " + Colors.FAIL + uniques_crashes_count + Colors.ENDC + "\r\n"
          + "Uniques timeouts: " + uniques_timeouts_count + "\r\n"
          + "Crashes: " + str(self.total_crashes) + "\r\n"
          + "Timeouts: " + total_timeouts + "\r\n"
          + "Reached paths: " + total_paths_count + "\r\n"
          + "Cycles: " + cycles_count + "\r\n"
          + "Executions: " + total_executions + "\r"
          )
    return True

def start_interface(self):
    """
    Display the curse interface
    """
    wrapper(self.init_interface)
    while True:
        self.refresh_interface()
        time.sleep(REFRESH_INTERVAL)


from frelatage.report.report import Report

def run_function(self, arguments: list, result: list) -> bool:
        """
        Run a function inside the tracer and return a report 
        """
        trace = self.tracer.trace(self.method, arguments)

        trace_instructions = set([instruction for instruction in trace.reached_instructions])
        trace_instruction_pairs = set([pair for pair in trace.instructions_pairs])

        trace_instructions_count = len(trace_instructions)
        trace_instruction_pairs_count = len(trace_instruction_pairs)

        new_instructions_count = len(trace_instructions - self.reached_instructions)
        new_sequences_count =  len(trace_instruction_pairs - self.instructions_pairs)

        new_instruction_error = trace.error_position not in self.error_positions

        run_report = Report(trace,
                            arguments, 
                            trace.error,
                            trace.timeout,
                            new_instruction_error,
                            trace_instructions_count,
                            new_instructions_count,
                            trace_instruction_pairs_count,
                            new_sequences_count)
        result.append(run_report)
        return True

def run_cycle(self) -> list[Report]:
    """
    run the function with each mutation of the cycle as argument.
    Return the execution reports.
    """
    cycle_reports = []
    
    # TODO: Implement multithreading
    for mutation in self.cycle:
        self.run_function(mutation, cycle_reports)
        self.inputs_count += 1
        self.stage_inputs_count += 1
    self.cycles_count += 1
    return cycle_reports


from frelatage.report.report import Report
import pickle
import os

def get_report_name(self, report: Report) -> str:
    """
    Generate a report title.
    The name of a report is in the following form:
    id:<crash ID>,err:<error type>,err_pos:<error>,err_file:<error file>,err_po:<err_pos>
    """
    # It is assumed that the number of uniques crashes will not exceed 999999
    error_id = str(self.unique_crashes).zfill(6) 
    # Type of the error in lowercase
    error_type = report.trace.error_type.lower()
    # Line of the file where the error occured
    error_position = str(report.trace.error_position[0][1]) if report.trace.error_position is not None else str(None)
    # File where the error occured
    error_file = os.path.splitext(os.path.basename(report.trace.error_position[0][0]))[0].lower()

    report_name = "id:{error_id},err:{error_type},err_file:{error_file},err_pos:{error_position}".format(
        error_id=error_id,
        error_type=error_type,
        error_file=error_file,
        error_position=error_position
    )
    return report_name

def save_report(self, report) -> bool:
    """
    Save a report in the output directory (/out by default).
    The arguments passed to the function and the behavior of the function are stored
    in "input", the file inputs are stored in files ranging from 0 to n.
    The report directory is in the following form: 
    ├── out
    │   ├── id:<crash ID>,err:<error type>,err_pos:<error>,err_file:<error file>,err_pos:<err_pos>
    │       ├── input
    │       ├── 0
    │           ├── <inputfile1>
    │       ├── ...
    │   ├── ...
    """

    # The report contains the parameters passed to the function.
    custom_report = {
        "input": [dict(input) for input in report.input]
    }

    # /out by default
    base_directory = self.output_directory
    report_name = self.get_report_name(report)

    report_directory = "{base_directory}/{report_name}".format(base_directory=base_directory, report_name=report_name)

    # Create ./out/<report directory>/ directory if not exists
    if not os.path.exists(report_directory):
        os.makedirs(report_directory)
    
    # Save report file
    with open('{report_directory}/input'.format(report_directory=report_directory), 'wb+') as f:
        # We use pickle to store the report object
        pickle.dump(custom_report, f)

    # Save input files
    for i in range(len(report.input)):
        argument = report.input[i]
        argument_number = i

        if argument.file:
            file_argument_content = open(argument.value, "rb").read()
            # Save file in /out/<report name>/<argument number>/<file name>
            argument_directory = "{report_directory}/{argument_number}".format(
                report_directory=report_directory,
                argument_number=argument_number
            )
            # create /out/<report name>/<argument number> folder if not exists
            if not os.path.exists(argument_directory):
                os.makedirs(argument_directory)
            
            filename = os.path.basename(argument.value)
            with open("{argument_directory}/{filename}".format(argument_directory=argument_directory, filename=filename), "wb+") as f:
                f.write(file_argument_content)
    return True


from typing import Any

class Report:

    def __init__(self,
                 trace: Any, 
                 input: list,
                 error: bool,
                 timeout: bool,
                 new_error_instruction: bool,
                 reached_instructions_count: int,
                 new_reached_instructions_count: int,
                 instructions_pairs_count: int,
                 new_instructions_pairs_count: int) -> None:
        """
        Report of the behavior of a function with given input
        """
        self.trace = trace
        self.input = input
        self.error = error
        self.timeout = timeout
        self.new_error_instruction = new_error_instruction
        self.reached_instructions_count = reached_instructions_count
        self.new_reached_instructions_count = new_reached_instructions_count
        self.instructions_pairs_count = instructions_pairs_count
        self.new_instructions_pairs_count = new_instructions_pairs_count
        
        self.score = self.compute_score()

    def compute_score(self) -> int:
        """
        The score of a report is used to select the most efficient inputs for the fuzzer's genetic algorithm mutations.
        The parameters are : the number of instructions reached, the number of new instructions reached, the number of
        pairs of instructions reached, the occurrence of an error/timeout.
        """
        
        result = self.reached_instructions_count + self.new_reached_instructions_count + self.instructions_pairs_count + self.new_instructions_pairs_count
        return result

    def __eq__(self, other):
        """
        Reports are sorted by score
        """
        return self.score == other.score

    def __lt__(self, other):
        """
        Reports are sorted by score
        """
        return self.score < other.score


from frelatage.tracer.trace_result import Result
from frelatage.config.config import Config
from typing import Callable, Iterable
import timeout_decorator
import os, contextlib
import functools
import trace
import time
import sys

def supress_stdout(func: Callable):
    """
    redirect the output of a function to /dev/null
    """
    def wrapper(*a, **ka):
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull):
                return func(*a, **ka)
    return wrapper

class Tracer:
    """
    The tracer allows to launch a function and to keep track of the executed instructions
    as well as the triggered errors.
    """
    def __init__(self,
                 exceptions_whitelist: Iterable[Exception] = [],
                 exceptions_blacklist: Iterable[Exception] = []
                 ) -> None:
        """
        Initialize the tracer with an exception whitelist and an exception blacklist.
        """
        self.exceptions_whitelist = exceptions_whitelist
        self.exceptions_blacklist = exceptions_blacklist

    @timeout_decorator.timeout(Config.FRELATAGE_TIMEOUT_DELAY, use_signals=True, timeout_exception=TimeoutError)
    def runfunc(self, function: Callable, arguments):
        """
        Trace a function.
        The result will be stored in self.tracer.
        """
        self.tracer.runfunc(function, *arguments)

    @supress_stdout
    def trace(self, function: Callable, arguments: list) -> Result:
        """
        Trace a function and return a trace result object.
        """
        error = False
        error_type = None
        timeout = False

        # Initialize the tracer
        self.tracer = trace.Trace(
            ignoredirs=[sys.prefix, sys.exec_prefix],
            trace=0,
            count=1
        )
        
        arguments = [argument.value for argument in arguments]

        # Run the function and catch errors
        try:
            self.runfunc(function, arguments)
        # Whitelisted exceptions
        except self.exceptions_whitelist as e:
            error = True
            error_type = str(e.__class__.__name__)
        # Timeout exception
        except TimeoutError as e:
            timeout = True
            error_type = str(e.__class__.__name__)
        # Blacklisted exceptions
        except self.exceptions_blacklist as e:
            pass
        # Default exceptions
        except Exception as e:
            if self.exceptions_whitelist == ():
                error = True
                error_type = str(e.__class__.__name__)
            else:
                pass

        # List of the executed instructions
        instructions = list(self.tracer.results().counter.items())
        reached_instructions = [(instruction[0][0], instruction[0][1]) for instruction in instructions]
        # List of the instructions pairs :
        # e.g : reached instructions : [('library.py', 1), ('library.py', 3), ('library.py', 2), ('library.py', 4)]
        #       instructions pairs -> [(('library.py', 3), ('library.py', 3)), (('library.py', 3), ('library.py', 2)), ...] 
        instructions_pairs = []
        for i in range(len(reached_instructions)):
            instructions_pairs.append((reached_instructions[i-1], reached_instructions[i]))
        # Position of the instruction where the error occured, None if no error
        if error or timeout:
            error_position = (reached_instructions[-1], error_type)
        else:
            error_position = None

        # Generate a result object
        result = Result(error, error_type, timeout, error_position, reached_instructions, instructions_pairs)
        return result


import re

class Dictionary:
    """
    Stores the elements of an Frelatage dictionary. 
    """
    def __init__(self) -> None:
        self.dictionary = []
        # dict_element="myelement" -> "myelement"
        self.DICTIONARY_ELEMENT_REGEXP = r'".+"'

    def load_dictionary_from_file(self, filename) -> list:
        """
        Read a Frelatage dictionary file and extract dictionary elements from it
        """
        file_dictionary = []

        with open(filename, "r") as f:
            lines = [line.replace("\n", "").strip() for line in f.readlines()]

            for line in lines:
                # Skip empty lines and commentaries
                if not line or line[0] == "#":
                    continue
                # Parse line
                dictionary_element = re.findall(self.DICTIONARY_ELEMENT_REGEXP, line)
                if dictionary_element:
                    # Remove the quote from both the beginning and the end
                    dictionary_element = dictionary_element[0][1:-1]
                    file_dictionary.append(dictionary_element)
        return file_dictionary


from frelatage.config.config import Config
from frelatage.input.input import Input 
import glob
import sys
import os

def load_corpus(directory: str, file_extensions: list = []) -> list[Input]:
    """
    Load a corpus from a file input directory subdirectory.
    Return a list of inputs.
    """
    inputs = []

    # ./<input directory>/<subdirectory>
    input_directory = os.path.join(
            os.path.dirname(os.path.realpath(sys.argv[0])),
            Config.FRELATAGE_INPUT_DIR,
            directory
    )

    # Search for file inputs in the subdirectory
    if not file_extensions:
        file_inputs = glob.glob(os.path.join(input_directory, "*"))
    # filter file inputs by extension
    else:
        file_inputs = []
        for file_extension in file_extensions:
            file_inputs += glob.glob(os.path.join(input_directory, "*.{extension}".format(extension=file_extension)))

    # Relative path
    file_inputs = [os.path.relpath(file_input, input_directory) for file_input in file_inputs]

    # Create an Input object for every file in the subdirectory
    for file_input in file_inputs:
        inputs.append(Input(file=True, value=file_input))
    return inputs


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['frelatage',
 'frelatage.config',
 'frelatage.corpus',
 'frelatage.dictionary',
 'frelatage.input',
 'frelatage.mutator',
 'frelatage.queue',
 'frelatage.report',
 'frelatage.tracer']

package_data = \
{'': ['*']}

install_requires = \
['timeout-decorator>=0.5.0,<0.6.0']

setup_kwargs = {
    'name': 'frelatage',
    'version': '0.0.5',
    'description': 'Frelatage is a coverage-based Python fuzzing library which can be used to fuzz python code.',
    'long_description': '<p align=center>\n  <img src="doc/frelatage_logo.gif" width="200" height="200" style="border-radius:4px"/>\n  <br>\n  <code>pip3 install frelatage</code></br>\n  <i>Current release : <a href="https://github.com/Rog3rSm1th/Frelatage/releases">0.0.5</a></i></br></br>\n  <a target="_blank" href="https://www.python.org/downloads/" title="Python version"><img src="https://img.shields.io/badge/Made%20with-Python-1f425f.svg"></a>\n  <a target="_blank" href="https://www.python.org/downloads/" title="Python version"><img src="https://img.shields.io/badge/python-%3E=_3.6-green.svg"></a>\n  <a target="_blank" href="LICENSE" title="License: MIT"><img src="https://img.shields.io/badge/License-MIT-blue.svg"></a>\n  <a target="_blank" title="Downloads"><img src="https://static.pepy.tech/badge/frelatage"></a>\n  <a target="_blank" href="https://twitter.com/Rog3rSm1th" title="Twitter"><img src="https://img.shields.io/badge/-@Rog3rSm1th-1ca0f1?style=flat-square&labelColor=1ca0f1&logo=twitter&logoColor=white&link=https://twitter.com/Rog3rSm1th"></a>\n  <br>\n  <span><i>The Python Fuzzer that the world deserves</i></span>\n</p>\n\n<p align="center">\n  <a href="#installation">Installation</a>\n  &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;\n  <a href="#how-it-works">How it works</a>\n  &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;\n  <a href="#features">Features</a>\n  &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;\n  <a href="#use-frelatage">Use Frelatage</a>\n  &nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;\n  <a href="#configuration">Configuration</a>\n</p>\n\n<p align="center">\n  <img src="https://github.com/Rog3rSm1th/Frelatage/blob/main/doc/frelatage_demo.gif?raw=true" alt="Frelatage demonstration"/>\n</p>\n\nFrelatage is a coverage-based Python fuzzing library which can be used to fuzz python code. The development of Frelatage was inspired by various other fuzzers, including [AFL](https://github.com/google/AFL)/[AFL++](https://github.com/AFLplusplus/AFLplusplus), [Atheris](https://github.com/google/atheris) and [PyFuzzer](https://github.com/eerimoq/pyfuzzer).The main purpose of the project is to take advantage of the best features of these fuzzers and gather them together into a new tool in order to efficiently fuzz python applications.\n\n**DISCLAIMER** : This project is at the alpha stage and can still cause many unexpected behaviors. Frelatage should not be used in a production environment at this time.\n\n## Requirements\n[Python 3](https://www.python.org/)\n\n## Installation\n\n#### Install with pip (recommended)\n\n```bash\npip3 install frelatage\n```\n\n#### Or build from source\n\nRecommended for developers. It automatically clones the main branch from the frelatage repo, and installs from source.\n\n```bash\n# Automatically clone the Frelatage repository and install Frelatage from source\nbash <(wget -q https://raw.githubusercontent.com/Rog3rSm1th/Frelatage/main/scripts/autoinstall.sh -O -)\n```\n\n## How it works\n\nThe idea behind the design of Frelatage is the usage of a genetic algorithm to generate mutations that will cover as much code as possible. The functioning of a fuzzing cycle can be roughly summarized with this diagram : \n\n```mermaid\ngraph TB\n\n    m1(Mutation 1) --> |input| function(Fuzzed function)\n    m2(Mutation 2) --> |input| function(Fuzzed function)\n    mplus(Mutation ...) --> |input| function(Fuzzed function)\n    mn(Mutation n) --> |input| function(Fuzzed function)\n    \n    function --> generate_reports(Generate reports)\n    generate_reports --> rank_reports(Rank reports)  \n    rank_reports --> select(Select n best reports)\n    \n    select --> |mutate| nm1(Mutation 1) & nm2(Mutation 2) & nmplus(Mutation ...) & nmn(Mutation n)\n    \n    subgraph Cycle mutations\n    direction LR\n    m1\n    m2\n    mplus\n    mn\n    end\n    \n    subgraph Next cycle mutations\n    direction LR\n    nm1\n    nm2\n    nmplus\n    nmn\n    end\n     \n    style function fill:#5388e8,stroke:white,stroke-width:4px\n```\n## Features\n\n#### Fuzzing different argument types: \n  - String\n  - Int\n  - Float\n  - List\n  - Tuple\n  - Dictionary\n  \n#### File fuzzing\nFrelatage allows to fuzz a function by passing a file as input. \n\n#### Fuzzer efficiency\n- Corpus\n- Dictionnary\n\n## Use Frelatage\n\n#### Fuzz a classical parameter\n\n```python\nimport frelatage\nimport my_vulnerable_library\n\ndef MyFunctionFuzz(data):\n  my_vulnerable_library.parse(data)\n\ninput = frelatage.Input(value="initial_value")\nf = frelatage.Fuzzer(MyFunctionFuzz, [[input]])\nf.fuzz()\n```\n\n#### Fuzz a file parameter\n\nFrelatage gives you the possibility to fuzz file type input parameters. To initialize the value of these files, you must create files in the input folder (```./in``` by default).\n\nIf we want to initialize the value of a file used to fuzz, we can do it like this:\n```bash\necho "initial value" > ./in/input.txt\n```\n\nAnd then run the fuzzer: \n\n```python\nimport frelatage\nimport my_vulnerable_library\n\ndef MyFunctionFuzz(data):\n  my_vulnerable_library.load_file(data)\n\ninput = frelatage.Input(file=True, value="input.txt")\nf = frelatage.Fuzzer(MyFunctionFuzz, [[input]])\nf.fuzz()\n```\n\n#### Load several files to a corpus at once\n\nIf you need to load several files into a corpus at once (useful if you use a large corpus) You can use the built-in function of Frelatage `load_corpus`. This function returns a list of inputs.\n\n```load_corpus(directory: str, file_extensions: list) -> list[Input]```\n- directory: Subdirectory of the input directory (relative path), e.g `./`, `./images`\n- file_extensions: List of file extensions to include in the corpus entries, e.g. `["jpeg", "gif"]`, `["pdf"]`\n\n```python\nimport frelatage\nimport my_vulnerable_library\n\ndef MyFunctionFuzz(data):\n  my_vulnerable_library.load_file(data)\n  my_vulnerable_library.load_file(data2)\n\n# Load every every file in the ./in directory\ncorpus_1 = frelatage.load_corpus(directory="./")\n# Load every .gif/.jpeg file in the ./in/images subdirectory\ncorpus_2 = frelatage.load_corpus(directory="./images", file_extension=["gif", "jpeg"])\n\nf = frelatage.Fuzzer(MyFunctionFuzz, [corpus_1, corpus_2])\nf.fuzz()\n```\n\n#### Fuzz with a dictionary\n\nYou can copy one or more dictionaries located [here](https://github.com/Rog3rSm1th/Frelatage/tree/main/dictionaries) in the directory dedicated to dictionaries (`./dict` by default).\n\n#### Examples\n\nYou can find more examples of fuzzers and corpus in the [examples directory](https://github.com/Rog3rSm1th/Frelatage/tree/main/examples).\n\n## Reports\n\nEach crash is saved in the output folder (```./out``` by default), in a folder named : ```id:<crash ID>,err:<error type>,err_pos:<error>,err_file:<error file>,err_pos:<err_pos>```.\n\nThe report directory is in the following form: \n```\n    ├── out\n    │   ├── id:<crash ID>,err:<error type>,err_pos:<error>,err_file:<error file>,err_pos:<err_pos>\n    │       ├── input\n    │       ├── 0\n    │            ├── <inputfile1>\n    │       ├── ...\n    │   ├── ...\n```\n\n#### Read a crash report\n\nInputs passed to a function are serialized using the [pickle](https://docs.python.org/3/library/pickle.html) module before being saved in the ```<report_folder>/input file```. It is therefore necessary to deserialize it to be able to read the contents of the file. This action can be performed with [this script](https://github.com/Rog3rSm1th/Frelatage/blob/main/scripts/read_report.py). \n\n```bash\n./read_report.py input\n```\n  \n## Configuration\n\nThere are two ways to set up Frelatage:\n\n#### Using the environment variables\n\n| ENV Variable                   | Description | Possible Values | Default Value |\n| -------------------------------| ----------- |--------|-------|\n| **FRELATAGE_DICTIONARY_ENABLE**   | Enable the use of mutations based on dictionary elements| ```1``` to enable, ```0``` otherwise | ```1``` |\n| **FRELATAGE_TIMEOUT_DELAY**        | Delay in seconds after which a function will return a TimeoutError | ```1``` - ```20``` | ```2``` |\n| **FRELATAGE_INPUT_FILE_TMP_DIR**   | Temporary folder where input files are stored | absolute path to a folder, e.g. ```/tmp/custom_dir```| ```/tmp/frelatage```|\n| **FRELATAGE_INPUT_MAX_LEN**        | Maximum size of an input variable in bytes | ```4``` - ```1000000``` |  ```4094``` |\n| **FRELATAGE_MAX_THREADS**          | Maximum number of simultaneous threads | ```8``` - ```50``` | ```8``` |\n| **FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS**      | Number of cycles without new paths found after which we go to the next stage | ```10``` - ```50000``` | ```5000``` | \n| **FRELATAGE_INPUT_DIR**           | Directory containing the initial input files. It needs to be a relative path (to the path of the fuzzing file) |relative path to a folder, e.g. ```./in```  | ```./in``` |\n| **FRELATAGE_DICTIONARY_DIR**      | Default directory for dictionaries. It needs to be a relative path (to the path of the fuzzing file) | relative path to a folder, e.g. ```./dict```  | ```./dict``` |  \n\nA configuration example :\n\n```bash\nexport FRELATAGE_DICTIONARY_ENABLE=1 &&\nexport FRELATAGE_TIMEOUT_DELAY=2 &&\nexport FRELATAGE_INPUT_FILE_TMP_DIR="/tmp/frelatage" &&\nexport FRELATAGE_INPUT_MAX_LEN=4096 &&\nexport FRELATAGE_MAX_THREADS=8 &&\nexport FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS=5000 &&\nexport FRELATAGE_INPUT_DIR="./in" &&\nexport FRELATAGE_DICTIONARY_DIR="./dict" &&\npython3 fuzzer.py\n```\n\n#### Passing arguments to the fuzzer \n\n```python\nimport frelatage \n\ndef myfunction(input1_string, input2_int):\n    pass\n\ninput1 = frelatage.Input(value="initial_value")\ninput2 = frelatage.Input(value=2)\n\nf = frelatage.Fuzzer(\n    # The method you want to fuzz\n    method=myfunction,\n    # Corpus\n    corpus=[[input1], [input2]],\n    # Number of threads\n    threads_count=8,\n    # Exceptions that will be taken into account\n    exceptions_whitelist=(OSError),\n    # Exceptions that will not be taken into account\n    exceptions_blacklist=(),\n    # Directory where the error reports will be stored\n    output_directory="./out",\n    # Enable or disable silent mode\n    silent=False\n)\nf.fuzz()\n```\n\n## Risks \n\nPlease keep in mind that, similarly to many other computationally-intensive\ntasks, fuzzing may put strain on your hardware and on the OS. In particular:\n\n  - Your CPU will run hot and will need adequate cooling. In most cases, if\n    cooling is insufficient or stops working properly, CPU speeds will be\n    automatically throttled. That said, especially when fuzzing on less\n    suitable hardware (laptops, smartphones, etc), it\'s not entirely impossible\n    for something to blow up.\n\n  - Targeted programs may end up erratically grabbing gigabytes of memory or\n    filling up disk space with junk files. Frelatage tries to enforce basic memory\n    limits, but can\'t prevent each and every possible mishap. The bottom line\n    is that you shouldn\'t be fuzzing on systems where the prospect of data loss\n    is not an acceptable risk.\n\n  - Fuzzing involves billions of reads and writes to the filesystem. On modern\n    systems, this will be usually heavily cached, resulting in fairly modest\n    "physical" I/O - but there are many factors that may alter this equation.\n    It is your responsibility to monitor for potential trouble; with very heavy\n    I/O, the lifespan of many HDDs and SSDs may be reduced.\n\n    A good way to monitor disk I/O on Linux is the \'iostat\' command:\n\n```shell\n    $ iostat -d 3 -x -k [...optional disk ID...]\n```\n\n## Contact \n\nfor any remark, suggestion, bug report, or if you found a bug using Frelatage, you can contact me at r0g3r5@protonmail.com or on twitter [@Rog3rSm1th](https://twitter.com/Rog3rSm1th)\n',
    'author': 'Rog3rSm1th',
    'author_email': 'r0g3r5@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Rog3rSm1th/Frelatage',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass

@dataclass
class Result:
    """
    Result of a trace
    """
    # True if an error occurred during the execution of a function, false otherwise
    error: bool
    # Error type, e.g TypeError, ZeroDivisionError...
    error_type: str
    # True if an TimeOutError occurred during the execution of a function, false otherwise
    timeout: bool
    # Position of the instruction where the error occured
    error_position: tuple
    # All groups of two successive instructions reached during the execution of a function
    # e.g : [(('library.py', 3), ('library.py', 3)), (('library.py', 3), ('library.py', 2)), ...] 
    instructions_pairs: list
    # List of the reached instructions during the execution of a function
    # e.g : [('library.py', 1), ('library.py', 3), ('library.py', 2), ('library.py', 4)]
    reached_instructions: list[tuple]


import sys
import os 

class FrelatageConfigError(Exception):
    pass

class Config:
    """
    We can pass configuration through environnement variables
    """
    # Enable the use of mutations based on dictionary elements
    if os.getenv('FRELATAGE_DICTIONARY_ENABLE', "1") not in ("1", "0"):
        raise FrelatageConfigError("FRELATAGE_DICTIONARY_ENABLE must '1' or '0'")
    FRELATAGE_DICTIONARY_ENABLE = True if os.getenv('FRELATAGE_DICTIONARY_ENABLE', "1") == "1" else False

    # Delay in seconds after which a function will return a timeoutError
    # Must be in range 1~20
    if not 1 <= int(os.getenv('FRELATAGE_TIMEOUT_DELAY', 2)) <= 20: 
        raise FrelatageConfigError("FRELATAGE_TIMEOUT_DELAY must be in range 1~20")
    FRELATAGE_TIMEOUT_DELAY = int(os.getenv('FRELATAGE_TIMEOUT_DELAY', 2))

    # Temporary folder where input files are stored. 
    # Need to be an absolute path
    FRELATAGE_INPUT_FILE_TMP_DIR = str(os.getenv('FRELATAGE_INPUT_FILE_TMP_DIR', '/tmp/frelatage'))

    # Maximum size of an input variable in bytes
    # Must be in range 4~1000000
    if not 4 <= int(os.getenv('FRELATAGE_INPUT_MAX_LEN', 4096)) <= 1000000:
        raise FrelatageConfigError("FRELATAGE_INPUT_MAX_LEN must be in range 4~1000000")
    FRELATAGE_INPUT_MAX_LEN = int(os.getenv('FRELATAGE_INPUT_MAX_LEN', 4096))

    # Maximum number of simultaneous threads
    # Must be in range 8~50
    if not 8 <= int(os.getenv('FRELATAGE_MAX_THREADS', 20)) <= 50:
        raise FrelatageConfigError("FRELATAGE_MAX_THREADS must be in range 8~50")
    FRELATAGE_MAX_THREADS = int(os.getenv('FRELATAGE_MAX_THREADS', 20))

    # Maximum number of successives cycle without a new path
    # Must be in range 10~50000
    if not 10 <= int(os.getenv('FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS', 5000)) <= 50000:
        raise FrelatageConfigError("FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS must be in range 10~10000")
    FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS = int(os.getenv('FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS', 5000))

    # Default directory for dictionaries
    # Relative path (to the path of the fuzzing file)
    FRELATAGE_DICTIONARY_DIR = str(os.path.join(
                                        os.path.dirname(os.path.realpath(sys.argv[0])),
                                        os.getenv('FRELATAGE_DICTIONARY_DIR', "./dict")
                                ))
    
    # Default directory for input files
    # Relative path (to the path of the fuzzing file)
    FRELATAGE_INPUT_DIR = str(os.path.join(
                                        os.path.dirname(os.path.realpath(sys.argv[0])),
                                        os.getenv('FRELATAGE_INPUT_DIR', "./in")
                          ))


from frelatage.dictionary.dictionary import Dictionary
import glob
import os

def load_dictionary(dictionary_folder: str) -> bool:
    """
    Load all the dictionaries from the dictionary folder
    """
    dictionary = Dictionary()
    dictionaries = []
    if os.path.isdir(dictionary_folder):
        # find all ".dict" files in the dictionary folder
        dictionary_files = glob.glob(os.path.join(dictionary_folder, "*.dict"))
        for file in dictionary_files:
            # parse dictionary files
            dictionaries += dictionary.load_dictionary_from_file(file)
    return dictionaries